/**
 * EasyBeans
 * Copyright (C) 2006-2008 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: DirectoryArchiveImpl.java 5101 2009-08-18 14:38:21Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.archive.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.jar.Manifest;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.archive.api.IArchiveMetadata;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import static org.ow2.util.url.URLUtils.fileToURL2;
import static org.ow2.util.url.URLUtils.fileToURL;
import org.ow2.util.url.URLUtilsException;

/**
 * Creates wrapper around directory.
 * @author Florent Benoit
 */
public class DirectoryArchiveImpl extends AbsArchiveImpl implements IArchive {

    /**
     * Manifest Path.
     */
    private static final String MANIFEST_PATH = "META-INF" + File.separator + "MANIFEST.MF";

    /**
     * Logger.
     */
    private static final Log logger = LogFactory.getLog(DirectoryArchiveImpl.class);

    /**
     * Internal resource used as archive.
     */
    private File directory = null;

    /**
     * URL of this directory.
     */
    private URL url = null;

    /**
     * Metadata analyzed ?
     */
    private boolean metadataAnalyzed = false;

    /**
     * Creates new instance of an IArchive for a directory.
     * @param directory the given directory.
     */
    protected DirectoryArchiveImpl(final File directory) {
        super();
        try {
            this.directory = directory.getCanonicalFile();
        } catch (IOException e) {
            logger.debug("Directory ''{0}'' cannot be canonicalized, use it as is", directory);
            this.directory = directory;
        }
        this.url = fileToURL(this.directory);
    }

    /**
     * @return a description of this archive. This name could be used in logger info.
     */
    public String getName() {
        return directory.getPath();
    }

    /**
     * @return the Metadata associated to this archive.
     */
    @Override
    public IArchiveMetadata getMetadata() {
        if (!metadataAnalyzed) {
            try {
                initMetadata();
            } catch (ArchiveException e) {
                logger.error("Cannot initialize metadata of file '" + directory + "'.", e);
            }
            metadataAnalyzed = true;
        }
        return super.getMetadata();

    }


    /**
     * Init metadata by reading the Manifest file.
     * @throws ArchiveException if metadata are not initialized
     */
    protected void initMetadata() throws ArchiveException {
        File manifestFile = new File(directory, MANIFEST_PATH);

        // Manifest exists, load it
        if (manifestFile.exists()) {
            Manifest manifest = null;
            InputStream is = null;
            try {
                try {
                    is = new FileInputStream(manifestFile);
                } catch (FileNotFoundException e) {
                    throw new ArchiveException("Cannot read the manifest file '" + manifestFile + "'", e);
                }
                try {
                    manifest = new Manifest(is);
                } catch (IOException e) {
                    throw new ArchiveException("Cannot read the manifest file '" + manifestFile + "'", e);
                }
            } finally {
                if (is != null) {
                    try {
                        is.close();
                    } catch (IOException e) {
                        throw new ArchiveException("Cannot close the manifest file '" + manifestFile + "'", e);
                    }
                }
            }

            // Fill metadata
            readManifest(manifest);
        }


    }

    /**
     * Encode resource name to be used on Unix/Windows systems.
     * @param resourceName the name to encode.
     * @return the encoded name.
     */
    private String encode(final String resourceName) {
        String[] tokens = resourceName.split("/");
        StringBuilder sb = new StringBuilder();
        for (String token : tokens) {
            if (sb.length() > 0) {
                sb.append(File.separator);
            }
            sb.append(token);
        }
        return sb.toString();
    }


    /**
     * Close the underlying Resource.
     * @return Returns <code>true</code> if the close was
     *         succesful, <code>false</code> otherwise.
     */
    public boolean close() {
        // Nothing to close
        return true;
    }

    /**
     * @param resourceName The resource name to be looked up.
     * @return Returns the resource URL if the resource
     *         has been found. null otherwise.
     * @throws ArchiveException if method fails.
     */
    public URL getResource(final String resourceName) throws ArchiveException {
        URL resourceURL = null;
        // lookup the directory on filesystem
        File f = new File(directory, encode(resourceName));
        try {
            if (f.exists()) {
                resourceURL = fileToURL2(f);
            }
        } catch (URLUtilsException e) {
            throw new ArchiveException("Invalid url", e);
        }
        return resourceURL;
    }

    /**
     * @return Returns an Iterator of Resource's URL.
     * @throws ArchiveException if method fails.
     */
    public Iterator<URL> getResources() throws ArchiveException {
        Map<String, URL> listResources = new HashMap<String, URL>();
        // Get all files and subdirectories
        addFiles(directory, listResources);
        return listResources.values().iterator();
    }

    /**
     * Methods that loop on directories to find the children (files).
     * @param file the given directory/file.
     * @param listResources the list on which to add new files.
     */
    private void addFiles(final File file, final Map<String, URL> listResources) {
        if (!file.exists() || !file.isDirectory()) {
            return;
        }

        // directory
        File[] files = file.listFiles();
        if (files != null) {
            for (File f : files) {
                // loop
                addFiles("", f, listResources);
            }
        }
    }

    /**
     * Methods that loop on directories to find the children (files).
     * @param name the resource name
     * @param file the given directory/file.
     * @param listResources the list on which to add new files.
     */
    private void addFiles(final String name, final File file, final Map<String, URL> listResources) {
        if (file.isDirectory()) {
            // directory
            File[] files = file.listFiles();
            if (files != null) {
                for (File f : files) {
                    // loop
                    addFiles(name + file.getName() + "/", f, listResources);
                }
            }
        } else {
            // single file
            try {
                listResources.put(name + file.getName(), fileToURL2(file));
            } catch (URLUtilsException e) {
                throw new IllegalStateException("Invalid url", e);
            }
        }
    }


    /**
     * @param resourceName The resource name to be looked up.
     * @return Returns an Iterator of matching resources.
     * @throws ArchiveException if method fails.
     */
    public Iterator<URL> getResources(final String resourceName) throws ArchiveException {
        List<URL> listResources = new ArrayList<URL>();
        File f = new File(directory, encode(resourceName));
        if (f.exists()) {
            try {
                listResources.add(fileToURL2(f));
            } catch (URLUtilsException e) {
                throw new ArchiveException("Invalid url", e);
            }
        }

        return listResources.iterator();

    }

    /**
     * @return Returns the resource URL.
     * @throws ArchiveException if method fails.
     */
    public URL getURL() throws ArchiveException {
        return this.url;
    }

    /**
     * Is that the given object is equals to our instance.
     * @param o the object to compare.
     * @return true if equals, else false.
     */
    @Override
    public boolean equals(final Object o) {
        if (!(o instanceof DirectoryArchiveImpl)) {
            return false;
        }
        DirectoryArchiveImpl other = (DirectoryArchiveImpl) o;
        return this.directory.equals(other.directory);
    }

    /**
     * Gets hashcode for this object.
     * @return hash code.
     */
    @Override
    public int hashCode() {
        return directory.hashCode();
    }

    /**
     * @return string representation
     */
    @Override
    public String toString() {
        return getName();
    }

    /**
     * @return all resources name
     */
    public Iterator<String> getEntries() {
        Map<String, URL> listResources = new HashMap<String, URL>();
        // Get all files and subdirectories
        addFiles(directory, listResources);
        return listResources.keySet().iterator();
    }
}
